package duckhuntbot;

import nintaco.api.*;
import static java.lang.Math.*;
import static duckhuntbot.Mode.*;

public class DuckHuntBot {
  
  private static enum State {
    WAITING_FOR_TARGET,
    ACQUIRING_TARGET,
    HOLDING_TRIGGER,
  }
  
  private static final int[] TARGET_ADDRESSES 
      = { 0x0300, 0x0350, 0x03A0, 0x04B0 };
  
  private static final int MODE_ADDRESS = 0x0676;
  private static final int TARGETS = TARGET_ADDRESSES.length;
  private static final int TARGET_SIZE = 32;
  private static final int DOG = 0x03;
  private static final int ACTIVE = 0x03;  
  private static final int ACQUIRING_TARGET_DELAY = 2;
  private static final int TRIGGER_HOLD_DELAY = 10;
  private static final int ACTIVE_Y0 = 8;
  private static final int ACTIVE_Y1 = 32;
  private static final int ACTIVE_Y2 = 144;
  private static final int ACTIVE_X = 80;
  private static final int BUSH_X1 = 192;
  private static final int BUSH_X2 = BUSH_X1 + 32;
  private static final int BUSH_Y = 120;
  
  private final Target[] targets = new Target[TARGETS];
  private final API api = ApiSource.getAPI();
  
  private State state = State.WAITING_FOR_TARGET;
  private int delay; 
  private Target holdingTarget;
  
  public void launch() {
    for(int i = TARGETS - 1; i >= 0; i--) {
      targets[i] = new Target();
    }
    api.addFrameListener(this::frameRendered);
    api.run();
  }
  
  private boolean readTargets(final int mode) {
    
    switch(mode) {
      case Initialized:
      case ShootingDucks:
      case ShootingClay:
      case BonusGame:
        break;
      default:
        return false;
    }
    
    for(int i = TARGETS - 1; i >= 0; i--) {
      final int address = TARGET_ADDRESSES[i];
      final Target target = targets[i];
      target.dog = api.readCPU(address) == DOG;
      target.active = api.readCPU(address + TargetAddressOffsets.State) 
          == ACTIVE;
      if (target.active) {
        target.y = api.readCPU(address + TargetAddressOffsets.Y);
        target.x = api.readCPU(address + TargetAddressOffsets.X);
        if (target.y < ACTIVE_Y0) {
          target.active = false;
        } else {        
          if (mode == ShootingDucks) {
            target.active = target.y <= ACTIVE_Y1 || (target.y <= ACTIVE_Y2 
              && target.x >= ACTIVE_X && (target.y < BUSH_Y 
                  || target.x < BUSH_X1 || target.x > BUSH_X2));
          } else {
            target.active = target.y < ACTIVE_Y2; 
          } 
        }
      }
    }
    
    // deactive dog and targets overlapping dog in bonus rounds
    for(int i = TARGETS - 1; i >= 0; i--) {
      final Target a = targets[i];
      if (a.dog && a.active) {
        a.active = false;
        for(int j = TARGETS - 1; j >= 0; j--) {
          if (i != j) {
            final Target b = targets[j];
            if (b.active && abs(a.x - b.x) <= TARGET_SIZE 
                && abs(a.y - b.y) <= TARGET_SIZE) {
              b.active = false;
            }
          }
        }
      }
    }
    
    for(int i = TARGETS - 1; i >= 0; i--) {
      if (targets[i].active) {
        return true;
      }
    }
    
    return false;
  }
  
  private void frameRendered() {
    
    final int mode = api.readCPU(MODE_ADDRESS);
    if (mode == PlayerSelect) {
      state = State.WAITING_FOR_TARGET;
      delay = 0;
      return;
    }
    
    switch(state) {
      case WAITING_FOR_TARGET:
        if (readTargets(mode)) {
          state = State.ACQUIRING_TARGET;
          delay = ACQUIRING_TARGET_DELAY;
        } else {
          api.setZapperTrigger(false);
        }
        break;
      case ACQUIRING_TARGET:
        outer: if (--delay <= 0) {
          if (readTargets(mode)) {
            for(int i = TARGETS - 1; i >= 0; i--) {
              final Target target = targets[i];
              if (target.active) {
                holdingTarget = target;
                api.setZapperY(target.y);
                api.setZapperX(target.x);
                api.setZapperTrigger(true);
                state = State.HOLDING_TRIGGER;
                delay = TRIGGER_HOLD_DELAY;
                break outer;
              }
            }
          } else {
            state = State.WAITING_FOR_TARGET;
            delay = 0;
          }
        }
        break;        
      case HOLDING_TRIGGER:
        readTargets(mode);
        api.setZapperY(holdingTarget.y);
        api.setZapperX(holdingTarget.x);
        api.setZapperTrigger(true);
        if (--delay <= 0) {
          state = State.WAITING_FOR_TARGET;
          delay = 0;
        }
        break;
    }
  }
  
  public static void main(final String... args) throws Throwable {
    System.out.println("Duck Hunt Bot started.");
    ApiSource.initRemoteAPI("localhost", 9999);
    new DuckHuntBot().launch();
  }
}